from RevFinder import *
import Metrics
import random
import Similarity
import time

'''Calculate the similarity scores of four similarity detection methods'''
def CRR(pastCommit_paths, newCommit_path, pastCommits, newCommit, pastReviewers, method):
    # recommendedReviewers = {}  # dict {reviewerName: score}
    scores = []
    # Calculate file path similarity
    if method == 'Jaccard coefficient':
        for pastFile in pastCommit_paths:
            scores.append(Similarity.jaccard_coefficient4str(newCommit_path, pastFile))
    if method == 'Hamming Similarity':
        for pastFile in pastCommit_paths:
            scores.append(Similarity.hammingDistance4str(newCommit_path, pastFile))
    # Calculate semantic similarity
    if method == 'RC_Cos_Similarity':
        for pastCommit in pastCommits:
            scores.append(Similarity.RC_Cos_Similarity(newCommit, pastCommit))
    if method == 'RC_Jaccard_Similarity':
        for pastCommit in pastCommits:
            scores.append(Similarity.jaccard_coefficient4text(newCommit, pastCommit))

    pastReviewers = dict(zip(pastReviewers, scores))

    recommendedReviewers = {k: v for k, v in sorted(pastReviewers.items(), key=lambda item: item[1], reverse=True)}  # reverse = True, descending order
    top_reviewers = {}
    # Return top 10 reviewers in the middle round, and then sum the different scores to get top-k reviewers
    for i, (key, value) in enumerate(recommendedReviewers.items()):
        if i in range(0, 10):
            top_reviewers[key] = value

    return top_reviewers


def MyCRRapproach(pastCommit_paths, newCommit_path, pastCommits, newCommit, pastReviewers, k):
    # freely choose and combine the selected similarity detection methods
    with concurrent.futures.ProcessPoolExecutor(max_workers=4) as executor:
        future1 = executor.submit(CRR, pastCommit_paths, newCommit_path, pastCommits, newCommit, pastReviewers, 'Jaccard coefficient')
        future2 = executor.submit(CRR, pastCommit_paths, newCommit_path, pastCommits, newCommit, pastReviewers, 'Hamming Similarity')
        future3 = executor.submit(CRR, pastCommit_paths, newCommit_path, pastCommits, newCommit, pastReviewers, 'RC_Cos_Similarity')
        future4 = executor.submit(CRR, pastCommit_paths, newCommit_path, pastCommits, newCommit, pastReviewers, 'RC_Jaccard_Similarity')

        scores_Jaccard = future1.result()
        scores_Hamming = future2.result()
        scores_RC_Cos = future3.result()
        scores_RC_Jaccard = future4.result()

        '''Multiple similarity detection methods'''
        Combined_score = {}
        for key in scores_Jaccard.keys() | scores_Hamming.keys() | scores_RC_Cos.keys() | scores_RC_Jaccard.keys():
            Combined_score[key] = sum(i.get(key, 0) for i in (scores_Jaccard, scores_Hamming, scores_RC_Cos, scores_RC_Jaccard))

    '''Multiple similarity detection methods'''
    Best_reviewers = {k: v for k, v in sorted(Combined_score.items(), key=lambda item: item[1], reverse=True)}
    top_k_reviewers = []
    for i, (key, value) in enumerate(Best_reviewers.items()):
        if i in range(0, k):
            top_k_reviewers.append(key)
    print('Top-%d_reviewer: ' % k, top_k_reviewers[:k])
    return top_k_reviewers[:k]

    # '''Individual similarity detection method'''
    # Best_reviewers = {k: v for k, v in sorted(scores_RC_Cos.items(), key=lambda item: item[1], reverse=True)}
    # top_k_reviewers = list(Best_reviewers.keys())[:k]
    # print('Top-%d_reviewer: ' % k, top_k_reviewers)
    # return top_k_reviewers


def RevFinder(pastCommit_paths, newCommit_paths, pastReviewers, Top_K):
    pre_Reviewers = recommendReviewers(pastCommit_paths, newCommit_paths, pastReviewers)
    # print('true_Reviewers: ', true_Reviewers)
    print('pre_Reviewers: ', pre_Reviewers[:Top_K])
    return pre_Reviewers[:Top_K]


def metrics(pastCommit_paths, newCommit_paths, pastCommits, newCommits, pastReviewers, true_Reviewers, method, k):
    precisions = []
    recalls = []
    is_Correct = []
    RRs = []

    for newCommit_path, newCommit, true_Reviewer in zip(newCommit_paths, newCommits, true_Reviewers):
        if method == 'RevFinder':
            pre_Reviewers = RevFinder(pastCommit_paths, newCommit_path, pastReviewers, Top_K)
        if method == 'MyCRRapproach':
            pre_Reviewers = MyCRRapproach(pastCommit_paths, newCommit_path, pastCommits, newCommit, pastReviewers, k)   # return a top-k list

        temp = []
        temp.append(true_Reviewer)
        print('(current) true_Reviewer: ', temp)
        precisions.append(Metrics.precisionK(temp, pre_Reviewers))
        recalls.append(Metrics.recallK(temp, pre_Reviewers))
        is_Correct.append(Metrics.isCorrect(temp, pre_Reviewers))
        RRs.append(Metrics.Reciprocal_rank(temp, pre_Reviewers))
    return precisions, recalls, is_Correct, RRs


def Evaluation(precisions, recalls, is_Correct, RRs, sample_number, k):
    # 实际的Mean precision
    Mean_precision_K = float(sum(precisions) / sample_number)
    Mean_recall_K = float(sum(recalls) / sample_number)
    TopK_accuracy = float((sum(is_Correct) / sample_number))
    MRR = float(sum(RRs) / sample_number)
    print("Mean_precision_%d: "%k, Mean_precision_K)
    print("Mean_recall_%d: "%k, Mean_recall_K)
    print("Top-%d_accuracy: "%k, TopK_accuracy)
    print("MRR: ", MRR)
    return Mean_precision_K, Mean_recall_K, TopK_accuracy, MRR


def data_split(full_list, ratio, shuffle=False):
    """
    :param full_list: data (a list)
    :param ratio: sampling percentage
    :param shuffle: whether shuffle the dataset
    :return:
    """
    n_total = len(full_list)
    sampling_number = int(n_total * ratio)
    if n_total == 0 or sampling_number < 1:
        print("An empty list!")
        return [], full_list
    # True: random sampling; False: get the last 10% sample as the test set
    if shuffle:
        random.seed(6)
        random.shuffle(full_list)
    training_set = full_list[:-sampling_number]
    test_set = full_list[-sampling_number:]
    return training_set, test_set


def Sampling_Evaluation(sampling_method, method, k):
    path = r'D:\dataset.xlsx'
    file = pd.read_excel(path, sheet_name='Neutron', na_values='n/a')
    # file = pd.read_excel(path, sheet_name='Nova', na_values='n/a')
    # file = pd.read_excel(path, sheet_name='QtBase', na_values='n/a')
    # file = pd.read_excel(path, sheet_name='QtCreator', na_values='n/a')

    '''incremental_sampling'''
    if sampling_method == 'incremental_sampling':
        initial_set = 29    # for the four projects: 29, 26, 23, 25
        step_size = 30      # for the four projects: 30, 45, 29, 7
        final_Mean_precision_K = []
        final_Mean_recall_K = []
        final_TopK_accuracy = []
        final_MRR = []

        # Initial set
        initial_sample_paths = file['File'][0:initial_set].tolist()          # [0, initial_set)
        initial_sample_Commits = file['Comment'][0:initial_set].tolist()
        initial_sample_Reviewers = file['Reviewer'][0:initial_set].tolist()
        # Iteration, test set (i.e., newXXX) and training set (i.e., pastXXX)
        initial_pastCommit_paths, initial_newCommit_paths = data_split(initial_sample_paths, 0.1, shuffle=False)
        initial_pastCommits, initial_newCommits = data_split(initial_sample_Commits, 0.1, shuffle=False)
        initial_pastReviewers, initial_true_Reviewers = data_split(initial_sample_Reviewers, 0.1, shuffle=False)
        initial_sample_size = len(initial_sample_paths)

        # Calculate the results on the initial set
        precisions, recalls, is_Correct, RRs = metrics(initial_pastCommit_paths, initial_newCommit_paths, initial_pastCommits, initial_newCommits, initial_pastReviewers, initial_true_Reviewers, method, k)
        Mean_precision_K, Mean_recall_K, TopK_accuracy, MRR = Evaluation(precisions, recalls, is_Correct, RRs, initial_sample_size, k)
        final_Mean_precision_K.append(Mean_precision_K)
        final_Mean_recall_K.append(Mean_recall_K)
        final_TopK_accuracy.append(TopK_accuracy)
        final_MRR.append(MRR)
        steps = int((len(file['File']) - initial_sample_size) / step_size)

        for step in range(1, steps + 1):        # [1, steps + 1)
            print('-------Step ', step, '-------')
            # fixed percentage，sample=10%
            sample_paths = file['File'][:(initial_sample_size + step * step_size)].tolist()
            sample_Commits = file['Comment'][:(initial_sample_size + step * step_size)].tolist()
            sample_Reviewers = file['Reviewer'][:(initial_sample_size + step * step_size)].tolist()
            sample_size = initial_sample_size + len(sample_paths)
            # Iteration, test set (i.e., newXXX) and training set (i.e., pastXXX)
            pastCommit_paths, newCommit_paths = data_split(sample_paths, 0.1, shuffle=False)
            pastCommits, newCommits = data_split(sample_Commits, 0.1, shuffle=False)
            pastReviewers, true_Reviewers = data_split(sample_Reviewers, 0.1, shuffle=False)

            # print("Length of Training set:", len(pastCommit_paths))
            # Metrics
            precisions, recalls, is_Correct, RRs = metrics(pastCommit_paths, newCommit_paths, pastCommits, newCommits, pastReviewers, true_Reviewers, method, k)
            Mean_precision_K, Mean_recall_K, TopK_accuracy, MRR = Evaluation(precisions, recalls, is_Correct, RRs, sample_size, k)
            print('****************')
            print('Current true_Reviewers list: ', true_Reviewers)
            print('Results of current step: Mean_precision_K, Mean_recall_K, TopK_accuracy, MRR')
            print(Mean_precision_K, Mean_recall_K, TopK_accuracy, MRR)
            print('****************')
            final_Mean_precision_K.append(Mean_precision_K)
            final_Mean_recall_K.append(Mean_recall_K)
            final_TopK_accuracy.append(TopK_accuracy)
            final_MRR.append(MRR)

        # Average metrics
        Mean_final_Mean_precision_K = float(sum(final_Mean_precision_K) / (1 + steps))
        Mean_final_Mean_recall_K = float(sum(final_Mean_recall_K) / (1 + steps))
        Mean_final_TopK_accuracy = float(sum(final_TopK_accuracy) / (1 + steps))
        Mean_final_MRR = float(sum(final_MRR) / (1 + steps))
        print('==============[Final Results]================')
        print("Mean_final_Mean_precision_%d: " % k, Mean_final_Mean_precision_K)
        print("Mean_final_Mean_recall_%d: " % k, Mean_final_Mean_recall_K)
        print("Mean_final_Top%d_accuracy: " % k, Mean_final_TopK_accuracy)
        print("Mean_final_MRR: ", Mean_final_MRR)


    if sampling_method == 'fixed_sampling':
        # (1) Fixed sampling number
        # sample_number = 10
        # newCommits_sample = file.sample(n=sample_number, axis=0, random_state=8)
        # (2) Fixed sampling percentage
        newCommits_sample = file.sample(axis=0, frac=0.1, random_state=8)  # sampling percentage: 10%
        sample_number = len(newCommits_sample)
        print("len(sample_number): ", sample_number)
        newCommit_paths = newCommits_sample['File'].tolist()
        newCommits = newCommits_sample['Comment'].tolist()
        true_Reviewers = newCommits_sample['Reviewer'].tolist()

        remaining_sample = file.drop(labels=newCommits_sample.index)
        pastCommit_paths = remaining_sample['File'].tolist()
        pastCommits = remaining_sample['Comment'].tolist()
        pastReviewers = remaining_sample['Reviewer'].tolist()

        precisions, recalls, is_Correct, RRs = metrics(pastCommit_paths, newCommit_paths, pastCommits, newCommits, pastReviewers, true_Reviewers, method, k)
        Mean_precision_K, Mean_recall_K, TopK_accuracy, MRR = Evaluation(precisions, recalls, is_Correct, RRs, sample_number, k)
        print('==============[Final Results]================')
        print("Mean_precision_%d: " % k, Mean_precision_K)
        print("Mean_recall_%d: " % k, Mean_recall_K)
        print("Top-%d_accuracy: " % k, TopK_accuracy)
        print("MRR: ", MRR)

    # return pastCommit_paths, newCommit_paths, pastCommits, newCommits, pastReviewers, true_Reviewers

if __name__ == '__main__':
    time_start = time.time()
    Top_K = 1       # 1, 3 ,5, 10
    # Choose a sampling technique and code reviewer recommendation approach
    Sampling_Evaluation('incremental_sampling', 'RevFinder', Top_K)
    # Sampling_Evaluation('incremental_sampling', 'MyCRRapproach', Top_K)
    # Sampling_Evaluation('fixed_sampling', 'RevFinder', Top_K)
    # Sampling_Evaluation('fixed_sampling', 'MyCRRapproach', Top_K)

    time_end = time.time()
    Evaluation_time = time_end - time_start
    print('-----------------------')
    print('Evaluation_time:', Evaluation_time, 's')