import re
import nltk
import string
import pandas as pd
import numpy as np
from nltk.stem import SnowballStemmer
from torchtext.data import get_tokenizer
from nltk.corpus import stopwords
from gensim.models.keyedvectors import KeyedVectors

'''Preprocessing'''
def recovery(tokens):
    recovey_sentence = ''
    for token in tokens:
        recovey_sentence = recovey_sentence + ' ' + token
    return recovey_sentence.strip()

def remove_url(text):
    text = text.replace('\r', '').replace('\n', '').replace('\t', '')
    text = re.sub(r'(https|http)?:\/\/(\w|\.|\/|\?|\=|\&|\%|-|\#)*\b(\/)?', '', text, flags=re.MULTILINE)
    text = re.sub(r'(\/.*?\/)+.*? ', '', text, flags=re.MULTILINE)
    text = re.sub(r'(www\.)\S+', '', text, flags=re.MULTILINE)
    text = re.sub(r'  ', ' ', text, flags=re.MULTILINE)
    text = text.replace('\r', '').replace('\n', '').replace('\t', '')
    return text

def preprocess(sentence):
    '''
    :param sentence: input a *string* (str type)
    :return: a token *list* (list type) of the string
    '''
    sentence = remove_url(sentence)
    tokenizer = get_tokenizer("basic_english")
    token_words = tokenizer(sentence)
    punctuation = list(set(string.punctuation))
    extra_special_characters = ["''", '``', '##','>>', 'e', 'g', 'eg', 'cant', 'cannot', 'isnt', 'would', 'could', 'doesnt', 'hasnt']
    special_characters = [c for c in extra_special_characters if c not in punctuation] + punctuation

    # Remove English stop words and special characters
    cleaned_words = [word for word in token_words if word not in stopwords.words('english')]
    cleaned_words = [word for word in cleaned_words if word not in special_characters]
    cleaned_words = [word for word in cleaned_words if not word.isdigit()]
    cleaned_words = [word for word in cleaned_words if word.encode('utf-8').isalpha()]
    # cleaned_words = [word.lower() for word in cleaned_words]

    return cleaned_words

'''Generate vectors'''
wv_from_bin = KeyedVectors.load_word2vec_format("D:\SO_vectors_200.bin", binary=True)
word2id = wv_from_bin.key_to_index  # dictionary, {word, index}
# read_path = r'D:\dataset.xlsx'
# data = pd.read_excel(read_path, sheet_name='Neutron', na_values='n/a')
# violation_comment = data['Comment'].tolist()


def get_word_vectors(embeding_model, word):
    if embeding_model == 'SO_w2v_200':
        if word == '0':
            word = '<UNK>'
        if word in word2id:
            vector = wv_from_bin.get_vector(word)
        else:
            vector = np.array([0.] * 200, dtype=np.float64)
    return vector

def get_sen_vectors(embeding_model, sentence):
    sentence_vec = np.array([0.] * 200, dtype=np.float64)
    for word in sentence:
        sentence_vec += get_word_vectors(embeding_model, word)
    return sentence_vec
